/***********************************************************************
** TxImage - Image Viewer
** Copyright (C) 2003
** Pierpaolo Di Panfilo <pippo_dp@libero.it>
** Cristian Di Panfilo <cridipan@virgilio.it>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.  To obtain a
** copy of the GNU Library General Public License, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** Any permitted reproduction of these routines, in whole or in part,
** must bear this legend.
**
*************************************************************************/


#include <qpe/resource.h>
#include <qpe/applnk.h>
#include <qapplication.h>
#include <qdialog.h>
#include <qlayout.h>
#include <qtoolbutton.h>
#include <qscrollview.h>
#include <qimage.h>
#include <qpixmap.h>
#include <qpopupmenu.h>
#include <qpalette.h>
#include <qsize.h>
#include <qframe.h>
#include <qlabel.h>
#include <qfileinfo.h>

#include "settings.h"
#include "imagereader.h"
#include "imagewidget.h"
#include "imageviewer.h"
#include "icons.h"

ZoomPopup::ZoomPopup(ImageViewer *imageViewer)
    : QPopupMenu(imageViewer)
{

    //zoom status label
    zoomLabel = new QLabel("Zoom", this);
    zoomLabel->setAlignment( AlignHCenter );
    zoomLabel->setFocusPolicy( QWidget::NoFocus );
    zoomLabel->setBackgroundColor( colorGroup().button().dark(100) );
    insertItem(zoomLabel);
    insertSeparator();

    //zoom out/in and max zoom
    QWidget *container = new QWidget(this);
    QHBoxLayout *hbox = new QHBoxLayout(container);
    zoomOut = new QToolButton(container);
    zoomOut->setPixmap( qembed_findPixmap("zoomout") );
    zoomOut->setAutoRaise( TRUE );
    zoomOut->setFocusPolicy( QWidget::TabFocus );
    connect( zoomOut, SIGNAL(clicked()), imageViewer, SLOT(zoomOut()) );
    hbox->addWidget(zoomOut);
    zoomIn = new QToolButton(container);
    zoomIn->setPixmap( qembed_findPixmap("zoomin") );
    zoomIn->setAutoRaise( TRUE );
    zoomIn->setFocusPolicy( QWidget::TabFocus );
    connect( zoomIn, SIGNAL(clicked()), imageViewer, SLOT(zoomIn()) );
    hbox->addWidget(zoomIn);
    zoomMax = new QToolButton(container);
    zoomMax->setPixmap( qembed_findPixmap("zoommax") );
    zoomMax->setAutoRaise( TRUE );
    zoomMax->setFocusPolicy( QWidget::TabFocus );
    connect( zoomMax, SIGNAL(clicked()), imageViewer, SLOT(zoomMax()) );
    hbox->addWidget(zoomMax);

    insertItem(container);
}

void ZoomPopup::setZoomStatus(QString status)
{
    zoomLabel->setText(status);
    zoomLabel->repaint();
}

void ZoomPopup::setZoom(int z)
{
    if(!z)
        zoomLabel->setText("Zoom");
    else

        zoomLabel->setText( QString("Zoom %1% ").arg(100/z) );
    zoomLabel->repaint();
}

void ZoomPopup::enableZoomIn(bool b)
{
    zoomIn->setEnabled(b);
}

void ZoomPopup::enableZoomOut(bool b)
{
    zoomOut->setEnabled(b);
}

void ZoomPopup::enableZoomMax(bool b)
{
    zoomMax->setEnabled(b);
}


ImageViewer::ImageViewer(QWidget  *parent, Settings *s, const char* name)
         : QDialog(parent, name, TRUE)
{
    settings = s;
    setCaption("Image Viewer");

    QVBoxLayout *vbox = new QVBoxLayout(this);

    //toolbar box
    QHBoxLayout *toolBox = new QHBoxLayout(vbox);

    QToolButton *infoB = new QToolButton(this);
    infoB->setPixmap( qembed_findPixmap("info") );
    infoB->setAutoRaise( TRUE );
    infoB->setFocusPolicy( QWidget::TabFocus );
    connect(infoB, SIGNAL(clicked()), SLOT(showInfoDialog()));
    toolBox->addWidget(infoB);
    toolBox->addStretch();

    statusLabel = new QLabel(this);
    statusLabel->setAlignment(AlignHCenter | AlignVCenter);
    toolBox->addWidget(statusLabel);
    toolBox->addStretch();

    QToolButton *scaleToScreenB = new QToolButton(this);
    scaleToScreenB->setPixmap( qembed_findPixmap("toscreen") );
    scaleToScreenB->setAutoRaise( TRUE );
    scaleToScreenB->setFocusPolicy( QWidget::TabFocus );
    connect(scaleToScreenB, SIGNAL(clicked()), SLOT(scaleToScreen()));
    toolBox->addWidget(scaleToScreenB);
    QToolButton *zoomB1 = new QToolButton(this);
    zoomB1->setPixmap( qembed_findPixmap("zoom") );
    zoomB1->setAutoRaise( TRUE );
    zoomB1->setFocusPolicy( QWidget::TabFocus );
    toolBox->addWidget(zoomB1);
    zoomPopup = new ZoomPopup( this );
    zoomB1->setPopup(zoomPopup);
    zoomB1->setPopupDelay(0);
    QToolButton *rotateB = new QToolButton(this);
    rotateB->setPixmap( qembed_findPixmap("rotate") );
    rotateB->setAutoRaise( TRUE );
    rotateB->setFocusPolicy( QWidget::TabFocus );
    toolBox->addWidget(rotateB);
    rotatePopup = new QPopupMenu(this);
    rotatePopup->insertItem( tr("Rotate 90"),this,
                       SLOT(rotate90()));
    rotatePopup->insertItem( tr("Rotate 180"),this,
                       SLOT(rotate180()));
    rotatePopup->insertItem( tr("Rotate 270"),this,
                       SLOT(rotate270()));
    rotatePopup->insertSeparator();
    rotatePopup->insertItem( tr("Rotate Best Fit"),this,
                       SLOT(rotateBestFit()), 0, AUTO_BEST_FIT);
    rotatePopup->setItemChecked(AUTO_BEST_FIT, settings->rotateBestFit() );
    rotateB->setPopup(rotatePopup);
    rotateB->setPopupDelay(0);

    QToolButton *brightB = new QToolButton(this);
    brightB->setPixmap( qembed_findPixmap("brightup") );
    brightB->setAutoRaise( TRUE );
    brightB->setFocusPolicy( QWidget::TabFocus );
    connect(brightB, SIGNAL(clicked()), SLOT(brightUpImage()));
    toolBox->addWidget(brightB);
    QToolButton *brightDownB = new QToolButton(this);
    brightDownB->setPixmap( qembed_findPixmap("brightdown") );
    brightDownB->setAutoRaise( TRUE );
    brightDownB->setFocusPolicy( QWidget::TabFocus );
    connect(brightDownB, SIGNAL(clicked()), SLOT(brightDownImage()));
    toolBox->addWidget(brightDownB);

    QToolButton *fullScreenB = new QToolButton(this);
    fullScreenB->setPixmap( Resource::loadPixmap( "fullscreen" ) );
    fullScreenB->setAutoRaise( TRUE );
    fullScreenB->setFocusPolicy( QWidget::TabFocus );
    connect(fullScreenB, SIGNAL(clicked()), SLOT(showFullScreen()));
    toolBox->addWidget(fullScreenB);

    scroll = new QScrollView(this, 0, WResizeNoErase|WNorthWestGravity);

    imageWid = new ImageWidget(scroll->viewport());
    imageWid->setPixmap(new QPixmap());
    scroll->addChild(imageWid);
    vbox->addWidget(scroll);

    brightness = 0;
    rotateDegree = 0;
    scaledToScreen = FALSE;

#if defined(Q_WS_QWS) || defined(_WS_QWS_)
    showMaximized();
#endif
}


void ImageViewer::reset()
{
    // clean up
    pm.resize(0,0);
    bpix.resize(0,0);
    imageWid->setPixmap(new QPixmap());
}

void ImageViewer::closeEvent( QCloseEvent *e) //'x' pressed
{
    reset();
    QDialog::closeEvent( e );
}

void ImageViewer::accept()  //ok pressed
{
    reset();
    QDialog::accept();
}

void ImageViewer::keyReleaseEvent( QKeyEvent *keyEvent)
{
    int key = keyEvent->key();
    switch(key)
    {
        case Key_Enter:
        case Key_Escape:
            reset();
            break;
        default:;
    }
    keyEvent->accept();
}

QSize ImageViewer::rectSize()
{
    // the size of the visible scroll
    QScrollBar *vBar = scroll->verticalScrollBar();
    QScrollBar *hBar = scroll->horizontalScrollBar();
    int bw = vBar->isVisible() ? vBar->width() : 0;
    int bh = hBar->isVisible() ? hBar->height() : 0;
    return QSize( scroll->visibleWidth()+bw, scroll->visibleHeight()+bh );
}

void ImageViewer::updateImage( QPixmap *pix )
{
    imageWid->setPixmap( pix );
    imageWid->repaint(FALSE);
}

void ImageViewer::load(DocLnk *d)
{
    doc = d;
    brightness = 0;
    rotateDegree = 0;
    scaledToScreen = FALSE;

    QFileInfo fi(doc->file());
    if(fi.size() > 50000) //loading takes time
        statusLabel->setText("Loading...");

    //on load by default scale the image to screen
    scaleToScreen();
    statusLabel->setText("");

    //calculate maximum zoom (the image can't be larger than 800x600
    //otherwise there will be a memory error showing up
    int i=9;
    while(i)
    {
        i--;
        if(!i)
            break;
        int width = imageWidth/i;
        int height = imageHeight/i;
        if((width*height)>(800*600))
            break;

    }
    max=i+1;
    qDebug("max zoom ratio=%d",max);

    //load image informations
    imageInfo[NAME] = doc->name();
    imageInfo[FORMAT] = QImage::imageFormat(doc->file())+QString(" image");
    imageInfo[LOCATION]=fi.dirPath();
    imageInfo[FILE_SIZE]=QString("%1 %2");
    //show file size in kb or bytes
    if(fi.size() > 1024 )
        imageInfo[FILE_SIZE]=imageInfo[FILE_SIZE].arg(fi.size()/1024).arg("Kb");
    else imageInfo[FILE_SIZE]=imageInfo[FILE_SIZE].arg(fi.size()).arg("bytes");

    imageInfo[DIMENSION] = QString("%1x%2").arg(imageWidth).arg(imageHeight);

}

void ImageViewer::showInfoDialog()
{
    QDialog *infoDialog = new QDialog(this, 0, TRUE);
    infoDialog->setCaption("Image Info");
    const char *fields[]={"Name","Format", "-", "Location","File Size", "-", "Dimension","Colors"};

    infoDialog->setMinimumSize(180,80);
    QGridLayout *gl = new QGridLayout(infoDialog, 6, 2, 4, 2);
    QLabel *l;
    int count=0;
    for ( int i=0; i<8; i++ )
    {
        if(fields[i]=="-")
        {
            QFrame *sep=new QFrame(infoDialog);
            sep->setFrameStyle(QFrame::HLine | QFrame::Sunken);
            gl->addMultiCellWidget(sep, i, i, 0, 1);

        }
        else
        {
            l = new QLabel(tr(fields[i]) + ":", infoDialog);
            gl->addWidget(l, i, 0);
            l = new QLabel(imageInfo[count], infoDialog);
            gl->addWidget(l, i, 1);
            count++;
        }
    }

    infoDialog->exec();
    delete infoDialog;
}


void ImageViewer::scaleToScreen()
{
    bool rotated=FALSE;

    if(scaledToScreen) //currently scaled to screen
        return;

    imageReader.load(doc->file());
    imageWidth = imageReader.imageWidth();
    imageHeight = imageReader.imageHeight();

    qApp->processEvents(); //it takes time
    if( settings->rotateBestFit() )
    {
        // get an image rotated for best fit
        // rotated get TRUE if the image has been rotated
        image = imageReader.getBestFitImage( rectSize(), rotated );
    }
    else
        image = imageReader.getImage( rectSize() );

     if ( image->numColors() > 0 )
        imageInfo[COLORS]=tr("%1 colors").arg(image->numColors());
    else if ( image->depth() >= 16 )
        imageInfo[COLORS]=tr("True color");

    pm.setOptimization( QPixmap::NormalOptim );
    pm.convertFromImage( *image );
    delete image;

    scaledToScreen=TRUE;

    if(rotated)
        rotateDegree = 90;
    if(rotateDegree && !rotated)
        rotate(0);

    if( brightness != 0 )
        brightImage();
    else
        updateImage(&pm);


    currentZoom=0; //no zoom
    rotatePopup->setItemEnabled(AUTO_BEST_FIT, TRUE);
    zoomPopup->setZoom(0); //no zoom
    zoomPopup->enableZoomIn(TRUE);
    zoomPopup->enableZoomOut(TRUE);
    zoomPopup->enableZoomMax(TRUE);
}

void ImageViewer::zoom(int ratio)
{
    if(!ratio)
        return;

    qDebug("zooming 1:%d",ratio);
    currentZoom = ratio;
    int width = imageWidth/currentZoom;
    int height = imageHeight/currentZoom;
    if((width*height)>(800*600))
        return;

    imageReader.load( doc->file() );
    image = imageReader.getImage( width, height );

    if(rotateDegree)
        image = imageReader.rotateImage(image, rotateDegree);

    pm.convertFromImage( *image );
    delete image;

    if( brightness != 0 )
        brightImage();
    else
        updateImage(&pm);

    scaledToScreen = FALSE;
    zoomPopup->setZoom(currentZoom);
    rotatePopup->setItemEnabled(AUTO_BEST_FIT, FALSE);
}

void ImageViewer::zoomIn()
{
    zoomPopup->setZoomStatus("Zooming in...");
    if(scaledToScreen)
    {
        int ratio=0;
        if(rotateDegree==90 || rotateDegree==270)
        {
            ratio = imageWidth/pm.height();
            ratio += imageWidth % pm.height() ? 1 : 0;
        }
        else
        {
            ratio = imageWidth/pm.width();
            ratio += imageWidth % pm.width() ? 1 : 0;
        }

        zoom(ratio-1);
    }else
        zoom(currentZoom-1);

    if(currentZoom==max)
    {
        zoomPopup->enableZoomIn(FALSE);
        zoomPopup->enableZoomMax(FALSE);
    }
}

void ImageViewer::zoomOut()
{
    zoomPopup->setZoomStatus("Zooming out...");
    if(scaledToScreen)
    {
        int ratio=0;
        if(rotateDegree==90 || rotateDegree==270)
            ratio = imageWidth/pm.height();
        else
            ratio = imageWidth/pm.width();

        zoom(ratio+1);
    }else
        zoom(currentZoom+1);

    zoomPopup->enableZoomIn(TRUE);
    zoomPopup->enableZoomMax(TRUE);
}

void ImageViewer::zoomMax()
{
    zoomPopup->setZoomStatus("Zooming...");
    zoom(max);
    zoomPopup->enableZoomIn(FALSE);
    zoomPopup->enableZoomMax(FALSE);
}

void ImageViewer::rotate(int degree)
{
    rotateDegree += degree;
    if(rotateDegree>=360) rotateDegree-=360;

    if(scaledToScreen)
    {
        imageReader.load(doc->file());
        image = imageReader.getRotatedImage( rectSize(), rotateDegree );
    }
    else //in zoom mode
    {
        image = new QImage(pm.convertToImage());
        pm.resize(0,0);
        image = imageReader.rotateImage(image, degree);
    }

    pm.convertFromImage( *image );
    if(image) delete image;

    if( brightness != 0 )
        brightImage();
    else
        updateImage( &pm );

    if(settings->rotateBestFit())
    {
        settings->setRotateBestFit( FALSE );
        rotatePopup->setItemChecked( AUTO_BEST_FIT, FALSE );
    }
}

void ImageViewer::rotate90()
{
    qDebug("rotate 90 clicked");
    rotate(90);
}


void ImageViewer::rotate180()
{
    qDebug("rotate 180 clicked");
    rotate(180);
}

void ImageViewer::rotate270()
{
    qDebug("rotate 270 clicked");
    rotate(270);
}

void ImageViewer::brightImage()
{
    bpix.resize(0,0);
    QImage *bImage = new QImage( pm.convertToImage() );
    intensity(bImage, (float)brightness/100);
    bpix.convertFromImage( *bImage );
    delete bImage;
    updateImage( &bpix );
}

void ImageViewer::brightUpImage()
{
    brightness += 10;
    qDebug("brightness =%d",brightness);
    brightImage();
}

void ImageViewer::brightDownImage()
{
    brightness -= 10;
    qDebug("brightness =%d",brightness);
    brightImage();
}

void ImageViewer::showFullScreen(DocLnk *d)
{
    ImageReader imageR;

    imageR.load( d->file() );
    QImage *img = imageR.getBestFitImage( qApp->desktop()->size() );
    if( brightness != 0 )
        intensity(img, (float)brightness/100);
    QPixmap *pix = new QPixmap();
    pix->convertFromImage(*img);
    imageR.reset();
    fullScreen = new ImageWidget(0, 0, WDestructiveClose);
    fullScreen->setBackgroundColor(Qt::black);
    fullScreen->setPixmap(pix, FALSE);
    fullScreen->resize(qApp->desktop()->size());
    if(isVisible())
    {
        hide();
        connect(fullScreen, SIGNAL(clicked()), SLOT(show()));
    }

    fullScreen->setFocus();
    fullScreen->showFullScreen();
    connect(fullScreen, SIGNAL(clicked()), fullScreen, SLOT(close()));
}

void ImageViewer::showFullScreen()
{
    showFullScreen(doc);
    connect(fullScreen, SIGNAL(CancelPressed()), SLOT(show()));
    connect(fullScreen, SIGNAL(clicked()), SLOT(show()));
}

void ImageViewer::rotateBestFit()
{
    bool rotated=FALSE;

    settings->setRotateBestFit( !settings->rotateBestFit() );
    rotatePopup->setItemChecked( AUTO_BEST_FIT, settings->rotateBestFit() );

    imageReader.load( doc->file() );
    if( settings->rotateBestFit() )
        //set rotated TRUE if the image has been rotated for best fit
        image = imageReader.getBestFitImage( rectSize(), rotated );
    else
        image = imageReader.getImage( rectSize() );

    pm.convertFromImage( *image );
    delete image;

    if(image)
    {
        if( brightness != 0 )
            brightImage();
        else
            updateImage(&pm);
    }

    if(rotated)
        rotateDegree=90;
    else
        rotateDegree=0;
}


// Intensity code courtesy of KDE project.

void ImageViewer::intensity(QImage *img, float percent)
{
    int segColors = img->depth() > 8 ? 256 : img->numColors();
    unsigned char *segTbl = new unsigned char[segColors];
    int pixels = img->depth() > 8 ? img->width()*img->height() :
                 img->numColors();
    unsigned int *data = img->depth() > 8 ? (unsigned int *)img->bits() :
                         (unsigned int *)img->colorTable();

    bool brighten = (percent >= 0);
    if ( percent < 0 )
        percent = -percent;

    if ( brighten )
    { // keep overflow check out of loops
        for ( int i=0; i < segColors; ++i )
        {
            int tmp = (int)(i*percent);
            if ( tmp > 255 )
                tmp = 255;
            segTbl[i] = tmp;
        }
    }
    else
    {
        for ( int i=0; i < segColors; ++i )
        {
            int tmp = (int)(i*percent);
            if ( tmp < 0 )
                tmp = 0;
            segTbl[i] = tmp;
        }
    }

    if ( brighten )
    { // same here
        for ( int i=0; i < pixels; ++i )
        {
            int r = qRed(data[i]);
            int g = qGreen(data[i]);
            int b = qBlue(data[i]);
            int a = qAlpha(data[i]);
            r = r + segTbl[r] > 255 ? 255 : r + segTbl[r];
            g = g + segTbl[g] > 255 ? 255 : g + segTbl[g];
            b = b + segTbl[b] > 255 ? 255 : b + segTbl[b];
            data[i] = qRgba(r, g, b,a);
        }
    }
    else
    {
        for ( int i=0; i < pixels; ++i )
        {
            int r = qRed(data[i]);
            int g = qGreen(data[i]);
            int b = qBlue(data[i]);
            int a = qAlpha(data[i]);
            r = r - segTbl[r] < 0 ? 0 : r - segTbl[r];
            g = g - segTbl[g] < 0 ? 0 : g - segTbl[g];
            b = b - segTbl[b] < 0 ? 0 : b - segTbl[b];
            data[i] = qRgba(r, g, b, a);
        }
    }
    delete [] segTbl;

}

