/***********************************************************************
** TxImage - Image Viewer
** Copyright (C) 2003
** Pierpaolo Di Panfilo <pippo_dp@libero.it>
** Cristian Di Panfilo <cridipan@virgilio.it>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.  To obtain a
** copy of the GNU Library General Public License, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** Any permitted reproduction of these routines, in whole or in part,
** must bear this legend.
**
*************************************************************************/

#include <qpe/applnk.h>
#include <qapplication.h>
#include <qimage.h>
#include <qpixmap.h>
#include <qhbox.h>
#include <qlabel.h>
#include <qtoolbutton.h>
#include <qfontmetrics.h>
#include <qtimer.h>
#include <qsize.h>

#include "settings.h"
#include "imagereader.h"
#include "imagewidget.h"
#include "slideshow.h"
#include "icons.h"


SlideShow::SlideShow(QList<DocLnk> *list, Settings *s)
         : ImageWidget(0, 0, WDestructiveClose )
{
    docList = list;
    settings = s;
    index = 0;
    last = FALSE;

    controlBar = new QHBox(this);
    controlBar->setFrameStyle( QFrame::StyledPanel | QFrame::Raised );

    QToolButton *back = new QToolButton(controlBar);
    back->setPixmap( qembed_findPixmap("back") );
    back->setAutoRaise(TRUE);
    connect(back, SIGNAL(clicked()), SLOT(showPreviousImage()));

    play = new QToolButton(controlBar);
    play->setPixmap( qembed_findPixmap("play") );
    play->setAutoRaise(TRUE);
    connect(play, SIGNAL(clicked()), SLOT(pause()));

    QToolButton *stop = new QToolButton(controlBar);
    stop->setPixmap( qembed_findPixmap("stop") );
    stop->setAutoRaise(TRUE);
    connect(stop, SIGNAL(clicked()), SLOT(close()));

    QToolButton *next = new QToolButton(controlBar);
    next->setPixmap( qembed_findPixmap("next") );
    next->setAutoRaise(TRUE);
    connect(next, SIGNAL(clicked()), SLOT(showNextImage()));

    counter = new QLabel("...", controlBar);
    counter->setMinimumWidth(35);
    counter->setAlignment( AlignVCenter | AlignCenter );

    resize(qApp->desktop()->size()); //set fullscreen
    setBackgroundColor(Qt::black);

    QSize size = controlBar->sizeHint();
    controlBar->setGeometry((width()-size.width())/2, height()-size.height()-5,
                            size.width(), size.height() );

    if( settings->controlBarInSlide() )
        controlBar->show();
    else controlBar->hide();

    slideTimer = new QTimer( this );
    connect( slideTimer, SIGNAL(timeout()),
             this, SLOT(showNextImage()) );
    connect( this, SIGNAL(clicked()), SLOT(showControlBar()));
}

SlideShow::~SlideShow()
{
    delete slideTimer;
}

void SlideShow::showControlBar()
{
    //show and hide control bar
    if(controlBar->isVisible())
        controlBar->hide();
    else controlBar->show();
}

void SlideShow::start()
{
    if(!docList->count())
        return;

    showFullScreen();
    setFocus();
    showNextImage();
}

void SlideShow::pause()
{
    if( slideTimer->isActive() ) //is playing
    {
        slideTimer->stop();
        play->setPixmap( qembed_findPixmap("play") );
    }
    else
    {
        play->setPixmap( qembed_findPixmap("pause") );
        showNextImage();
    }
}

void SlideShow::stop()
{
    slideTimer->stop();
    index=0;
    close();
}

void SlideShow::showImage( QString file )
{
    QImage img;

    qApp->processEvents();
    imageReader.load( file );
    if( settings->rotateInSlide() )
        img = *imageReader.getBestFitImage( qApp->desktop()->size() );
    else
        img = *imageReader.getImage( qApp->desktop()->size() );

    pm.resize(0,0); //reset previous pixmap
    pm.convertFromImage( img );
    imageReader.reset();
    setPixmap( &pm, FALSE ); // not resize
    repaint();
    updateCounter(); //update counter in control bar
}

void SlideShow::showNextImage()
{
    if(last)
        close();

    DocLnk *doc = docList->at(index);
    if(doc && !closing)
    {
        showImage( doc->file() );
        slideTimer->start( settings->slideDelay() * 1000 , TRUE );
        if( index == (docList->count() - 1) ) //is last
        {
            index=0;
            last = TRUE;
        }
    }
    index++;
}

void SlideShow::showPreviousImage()
{
    if(!index) //first slide
        return;

    if(slideTimer->isActive())
        pause();

    DocLnk *doc = docList->at(--index);
    if(doc && !closing)
        showImage( doc->file() );
}

void SlideShow::updateCounter()
{
    QString count = QString("%1/%2").arg(index+1).arg(docList->count());
    counter->setText( count );
}

void SlideShow::keyReleaseEvent( QKeyEvent *keyEvent)
{
    int key = keyEvent->key();
    switch(key)
    {
        case Key_Enter:
        case Key_Down:
        case Key_Right:
            slideTimer->stop();
            showNextImage();
            break;
        case Key_Left:
        case Key_Up:
            slideTimer->stop();
            showPreviousImage();
            break;
        case Key_Space:
            pause();
            break;
        case Key_Escape:
            stop();
            break;
        default:;
    }
    keyEvent->accept();
}
