/***********************************************************************
** TxImage - Image Viewer
** Copyright (C) 2003
** Pierpaolo Di Panfilo <pippo_dp@libero.it>
** Cristian Di Panfilo <cridipan@virgilio.it>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.  To obtain a
** copy of the GNU Library General Public License, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** Any permitted reproduction of these routines, in whole or in part,
** must bear this legend.
**
*************************************************************************/


#include <qpe/applnk.h>
#include <qscrollview.h>
#include <qpalette.h>
#include <qapplication.h>
#include <qlayout.h>
#include <qfontmetrics.h>
#include <qabstractlayout.h>
#include <qlist.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qdir.h>
#include <string.h>
#include <qimage.h>

#include "imageviewer.h"
#include "settings.h"
#include "imagereader.h"
#include "imagelistview.h"
#include "tximage.h"
#include "thumbnailview.h"
#include "icons.h"

#define THUMB_WIDTH 80
#define THUMB_HEIGHT 60


ImageBox::ImageBox(QWidget *parent, Settings *s, ImageListViewItem *i, const char* name)
        : QWidget(parent, name, WResizeNoErase | WRepaintNoErase )
{
    //show this pixmap until the thumbnail is loaded
    scaledPixmap = new QPixmap( qembed_findPixmap("image") );
    //memory optimization for pixmap smaller than 128 px
    scaledPixmap->setOptimization( QPixmap::MemoryOptim );

    doc = i->docLnk();
    settings = s;
    item = i;
    imgWidth=0;
    imgHeight=0;
    selected = FALSE;
    thumbLoaded = FALSE;

    //get the file name with the extension
    int pos = doc->file().findRev('/');
    fileName = doc->file().right( doc->file().length()-pos-1 );

    setBackgroundMode( NoBackground );
    setFocusPolicy( QWidget::WheelFocus ); //get focus by clicking and keyboard
    setMinimumSize( THUMB_WIDTH+20, THUMB_HEIGHT+35 ); //margins
}

ImageBox::~ImageBox()
{
    delete scaledPixmap; //delete thumbnail
}

void ImageBox::setSelected(bool sel)
{
    //select the image box and get focus.
    //it remain selected even if it lose focus
    selected = sel;
    if(sel) setFocus();
}

bool ImageBox::isSelected()
{
    return selected;
}

void ImageBox::paintEvent( QPaintEvent *e )
{
    int pWidth = THUMB_WIDTH, pHeight = THUMB_HEIGHT;
    if(thumbLoaded)
    {
        //resize imagebox to thumbnail size
        pWidth = scaledPixmap->width();
        pHeight = scaledPixmap->height();
    }
    int left = (width() - pWidth) / 2;
    int right = left + pWidth +1;
    int top = ((height() - 30 - pHeight) / 2) + 3;
    int bottom = top + pHeight + 1;

    QPainter painter(this);
    painter.setClipRect(e->rect());
    painter.fillRect( rect(), isSelected() ? colorGroup().highlight() : colorGroup().base());
    if(thumbLoaded) //draw the thumbnail
        painter.drawPixmap( left+1, top+1, *scaledPixmap);
    else //draw the image icon centered
        painter.drawPixmap( (width() - scaledPixmap->width())/2,
                             ((height() - 30 - scaledPixmap->height()) / 2) + 3,
                             *scaledPixmap );

    //draw rounded borders
    painter.setPen(Qt::black);
    painter.drawLine(left+2, bottom, right-2, bottom);
    painter.drawLine(right, top+2, right, bottom-2);
    painter.drawPoint(right-1, bottom-1);
    painter.drawPoint(right-1, top+1);
    painter.setPen(Qt::lightGray);
    painter.drawLine(left+2, top, right-2, top);
    painter.drawLine(left, top+2, left, bottom-2);
    painter.drawPoint(left+1, top+1);
    painter.drawPoint(left+1, bottom-1);

    painter.setPen( isSelected() ? colorGroup().highlightedText() : colorGroup().text());

    QString info = settings->showFileExtension() ? fileName : doc->name();
    //justify image name
    QString name = info;
    QFontMetrics fm = painter.fontMetrics();
    int nameWidth = fm.width(info);
    if(nameWidth > width())
    {
        QString nameJustify = "...";
        int i = 0;
        while ( fm.width( nameJustify + info[ i ] ) < width() )
            nameJustify += info[ i++ ];
        nameJustify.remove( 0, 3 );
        if ( nameJustify.isEmpty() )
            nameJustify = info.left( 1 );
        nameJustify += "...";
        info = nameJustify;
    }

    //add image dimension
    if( settings->showImageDimension() && imgWidth && imgHeight)
        info += QString("\n%1x%2").arg(imgWidth).arg(imgHeight);

    painter.drawText(0,height()-27,width(),27, AlignCenter, /*name*/info);

}

void ImageBox::mouseReleaseEvent(QMouseEvent* )
{
    emit clicked();
}

void ImageBox::keyPressEvent( QKeyEvent *e )
{
    //signals for thumbnailview to focus and select next/previous images
    switch ( e->key() ) {
        case Key_Space:
            emit clicked();
            break;
        case Key_Up:
            emit upPressed();
            break;
        case Key_Left:
            emit leftPressed();
            break;
        case Key_Down:
            emit downPressed();
            break;
        case Key_Right:
            emit rightPressed();
            break;

        default: ;
    }
}

void ImageBox::loadThumb()
{
    bool saveThumbs = settings->saveThumbnail();
    bool loadedSavedThumb=FALSE;
    ImageReader imageReader;
    QString location, thumbFile, thumbpath;

    if( saveThumbs )
    {
        int pos = doc->file().findRev('/');
        location = doc->file().left(pos);
        thumbpath = QString( location+"/.txthumbs/");
        thumbFile = thumbpath+fileName;
    }

    qApp->processEvents();

    QImage *image = new QImage();
    if( saveThumbs && QFile::exists(thumbFile) )
    {
       //load saved thumbnail
        QFile f(thumbFile);
        f.open( IO_ReadOnly);
        char buf[12];
        //the first line contains the original image dimension
        f.readLine( buf, 12);
	sscanf(buf, "%d %d", &imgWidth, &imgHeight);
        QImageIO iio(&f, "BMP");
        iio.setImage( *image );
        iio.read();
        *image = iio.image();
        f.close();
        loadedSavedThumb=TRUE;
    }
    else
    {
       //load the image
        imageReader.load(doc->file());
        imgWidth = imageReader.imageWidth();
        imgHeight = imageReader.imageHeight();
        //smooth scale
        qDebug("loading thumbnail " + doc->name() );
        image = imageReader.getImage(80, 60);
    }

    //set image dimension in listviewitem
    item->setText(1, QString("%1x%2").arg(imgWidth).arg(imgHeight));

    if( image && scaledPixmap->convertFromImage( *image ) ) //conversion successfull
    {
        if(saveThumbs && !loadedSavedThumb )
        {
            QFileInfo fi(doc->file());
            //save thumb only for big images
            if( fi.size() > (uint)settings->thumbFileSize()*1024)
            {
                if( !QFile::exists(thumbpath) )
                {
                    //create the thumbnail directory (.txthumb)
                    QDir dir(location);
                    if ( !dir.mkdir(".txthumbs", FALSE) )
                        qWarning("can't create directory: "+location+"/.txthumbs");
                }
                if( QFile::exists(thumbpath) )
                {
                   //save thumbnail
                   //write a line with image size and then the bitmap
                    QFile f(thumbFile);
                    f.open( IO_WriteOnly );
                    QString res = QString("%1 %2\n").arg(imgWidth).arg(imgHeight);
                    f.writeBlock( res.data(), res.length() );
                    QImageIO iio(&f, "BMP");
                    iio.setImage( *image );
                    iio.write();
                 }
            }
        }

        delete image;
        thumbLoaded = TRUE;
        repaint(FALSE); //repaint no erase
    }

}

DocLnk *ImageBox::docLnk()
{
    return doc;
}


ThumbnailView::ThumbnailView(QWidget *parent, Settings *s, const char *name)
             : QScrollView(parent, name)
{
    boxList.setAutoDelete(TRUE);
    settings = s;
    enableClipper(TRUE); //optimization for more widget
    viewport()->setBackgroundColor(colorGroup().base());
    container = new QWidget(this, 0, WResizeNoErase | WRepaintNoErase | WPaintClever);
    container->setBackgroundColor(colorGroup().base());
    addChild(container);
    gridLayout = new QGridLayout(container, 1, 2, 4, 8);
}

ThumbnailView::~ThumbnailView()
{
    boxList.clear();
}

void ThumbnailView::setFocus()
{
    if( !boxList.count() ) //no images
        return;

    ImageBox *box = selectedImage();
    if(box)
        box->setFocus();
    else
    {
        box = boxList.at(0); //first image
        box->setSelected(TRUE);
    }
}

void ThumbnailView::addImage(ImageListViewItem *d, int row, int col)
{
    ImageBox *box = new ImageBox(container, settings, d);
    connect(box, SIGNAL(clicked()), SLOT(boxClicked()));
    connect(box, SIGNAL(rightPressed()), SLOT(rightPressed()));
    connect(box, SIGNAL(leftPressed()), SLOT(leftPressed()));
    connect(box, SIGNAL(upPressed()), SLOT(upPressed()));
    connect(box, SIGNAL(downPressed()), SLOT(downPressed()));
    gridLayout->addWidget(box, row, col);
    boxList.append(box);
    box->show();
}


void ThumbnailView::boxClicked()
{
    ImageBox *box = selectedImage();
    if(box) //deselect old image
        box->setSelected(FALSE);

    //select clicked image
    box = (ImageBox *)sender();
    box->setSelected(TRUE);
    DocLnk *doc = box->docLnk();
    emit imageClicked(doc);
}

void ThumbnailView::rightPressed()
{
    //select and focus next image
    ImageBox *box = (ImageBox*)sender();
    int pos = boxList.findRef(box);
    if(pos==-1)
        return;
    box->setSelected(FALSE);
    selectImage(pos+1);
}

void ThumbnailView::leftPressed()
{
    //select and focus previous image
    ImageBox *box = (ImageBox*)sender();
    int pos = boxList.findRef(box);
    if(pos==-1)
        return;
    box->setSelected(FALSE);
    selectImage(pos-1);
}

void ThumbnailView::upPressed()
{
    //select and focus up image
    ImageBox *box = (ImageBox*)sender();
    int pos = boxList.findRef(box);
    if(pos==-1)
        return;
    box->setSelected(FALSE);
    selectImage(pos-2);
}

void ThumbnailView::downPressed()
{
    //select and focus down image
    ImageBox *box = (ImageBox*)sender();
    int pos = boxList.findRef(box);
    if(pos==-1)
        return;
    box->setSelected(FALSE);
    selectImage(pos+2);
}

void ThumbnailView::selectImage(int pos)
{
    if( pos < 0 || pos > (int)(boxList.count()-1) )
        return;

    ImageBox *box = boxList.at(pos);
    if(box)
    {
        box->setSelected(TRUE);
        //scroll to show the selected image box
        ensureVisible(box->x(), box->y(), box->width(), box->height());
    }
}

ImageBox *ThumbnailView::find(DocLnk *d)
{
    QListIterator<ImageBox> it(boxList);
    for( ; it.current(); ++it)
    {
        ImageBox *box = it.current();
        if( box->docLnk() == d )
            return box;
    }
    return 0;
}

ImageBox *ThumbnailView::selectedImage()
{
    QListIterator<ImageBox> it(boxList);
    for( ; it.current(); ++it)
    {
        ImageBox *box = it.current();
        if( box->isSelected() )
            return box;
    }
    return 0;
}

void ThumbnailView::loadThumbs()
{
    //load all thumbnails
    QListIterator<ImageBox> it(boxList);
    for( ; it.current();++it)
    {
        ImageBox *box = *it;
        box->loadThumb();
    }
    qApp->processEvents();
}

void ThumbnailView::deleteImage(ImageBox *box)
{
    int row=0, col=0;

    //delete box
    boxList.remove(box);
    clear();
    //reload all to fill blank space
    QListIterator<ImageBox> it(boxList);
    for( ; it.current(); ++it) {
        ImageBox *box = *it;
        gridLayout->addWidget(box, row , col);
        col++;
        if(col==2) {
           row++;
           col=0;
        }
     }
}

void ThumbnailView::deleteImage(DocLnk *d)
{
    deleteImage( find(d) );
}

void ThumbnailView::clear(bool del)
{
    QLayoutIterator it(gridLayout->iterator());
    while( it.current() )
        del ? (void)it.deleteCurrent() : (void)it.takeCurrent();
    if(del) boxList.clear();
}
