/***********************************************************************
** TxImage - Image Viewer
** Copyright (C) 2003
** Pierpaolo Di Panfilo <pippo_dp@libero.it>
** Cristian Di Panfilo <cridipan@virgilio.it>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.  To obtain a
** copy of the GNU Library General Public License, write to the Free
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** Any permitted reproduction of these routines, in whole or in part,
** must bear this legend.
**
*************************************************************************/


#include <qpe/qpeapplication.h>
#include <qpe/qpetoolbar.h>
#include <qpe/resource.h>
#include <qpe/applnk.h>
#include <qpe/global.h>
#include <qpe/config.h>
#include <qmainwindow.h>
#include <qlayout.h>
#include <qtoolbutton.h>
#include <qaction.h>
#include <qpopupmenu.h>
#include <qscrollview.h>
#include <qtimer.h>
#include <qwidgetstack.h>
#include <qstring.h>
#include <qlist.h>
#include <qpalette.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qmessagebox.h>
#include <qdir.h>
#include <qcopchannel_qws.h>

#include "tximage.h"
#include "thumbnailview.h"
#include "imagelistview.h"
#include "imageviewer.h"
#include "slideshow.h"
#include "bookmark.h"
#include "settings.h"
#include "settingsdialog.h"
#include "icons.h"

TxImage::TxImage(QWidget *parent, const char *name, int flags)
           : QMainWindow(parent, name, flags)
{
    setCaption( tr("Tx Image Viewer 0.2") );
    setIcon( qembed_findPixmap("tximage") );

    /* if the width of the pixmap is less than or equal to 128 pixels,
    Qt stores the pixmap in a way which is very memory-efficient */
    QPixmap::setDefaultOptimization(QPixmap::MemoryOptim);

    setRightJustification(TRUE);
    mToolbar = new QPEToolBar(this);
    mToolbar->setHorizontalStretchable(TRUE);
    setToolBarsMovable(FALSE);

    //main toolbar
    bookmarkB = new QToolButton( qembed_findPixmap("bookmark_folder"),
                                   "Bookmarks", 0, this, SLOT(showBookmarkMenu()),
                                   mToolbar);
    mToolbar->addSeparator();
    thumbViewB = new QToolButton( qembed_findPixmap("view_icon"), 0, 0,
                                 this, SLOT(switchThumbnailView()), mToolbar);
    thumbViewB->setToggleButton(TRUE);
    lView = new QToolButton( qembed_findPixmap("view_text"), 0, 0,
                                 this, SLOT(switchListView()), mToolbar);
    lView->setToggleButton(TRUE);
    QAction *slide = new QAction("SlideShow", qembed_findPixmap("slideshow"), 0,0,this);
    connect(slide, SIGNAL(activated()),this,SLOT(startSlideShow()));
    slide->addTo(mToolbar);

    QToolButton *settingsB = new QToolButton( Resource::loadPixmap("SettingsIcon"),
                                            0, 0, this, 0, mToolbar);
    settingsPopup = new QPopupMenu(this);
    settingsPopup->insertItem( tr("Settings"), this, SLOT(showSettingsDialog()));
    settingsPopup->insertSeparator();
    settingsPopup->insertItem( tr("Show Images Fullscreen"),this,
                       SLOT(setFullScreen()), 0, FULLSCREEN );
    settingsB->setPopup(settingsPopup);
    settingsB->setPopupDelay(0);

    //add space
    QWidget *stretchWidget = new QWidget(mToolbar);
    stretchWidget->setBackgroundColor(colorGroup().button());
    mToolbar->setStretchableWidget(stretchWidget);

    QAction *fullScreen = new QAction("Show FullScreen",
                                   Resource::loadPixmap("fullscreen"),0,0,this);
    connect(fullScreen, SIGNAL(activated()),this,SLOT(showImageFullScreen()));
    fullScreen->addTo(mToolbar);
    QAction *rename = new QAction("Rename", qembed_findPixmap("rename"), 0,0,this);
    connect(rename, SIGNAL(activated()),this,SLOT(renameImage()));
    rename->addTo(mToolbar);
    QAction *deleteB = new QAction("delete", Resource::loadPixmap("editdelete"),
                                  0,0,this);
    connect(deleteB, SIGNAL(activated()),this,SLOT(deleteImage()));
    deleteB->addTo(mToolbar);

    settings = new Settings();
    settings->readSettings();

    //image viewer
    imageViewer = new ImageViewer(this, settings);
    imageViewer->hide();

    stack = new QWidgetStack(this);
    setCentralWidget(stack);

    //thumbnail view
    thumbView = new ThumbnailView(stack, settings);
    connect(thumbView, SIGNAL(imageClicked(DocLnk *)), SLOT(showImage(DocLnk *)));
    stack->addWidget(thumbView, THUMBNAIL_VIEW);

    //listview
    listView = new ImageListView(stack);
    connect(listView, SIGNAL(clicked(QListViewItem*)),
          this, SLOT(listItemClicked(QListViewItem*)));
    stack->addWidget(listView, LIST_VIEW);

    //use memory optimization only for toolbars
    QPixmap::setDefaultOptimization(QPixmap::NormalOptim);

    settingsPopup->setItemChecked(FULLSCREEN, settings->showImageFullScreen());

    //receive a signal when a card is mounted
    QCopChannel *channel = new QCopChannel("QPE/Card", this);
    bookmarkMenu = new BookmarkMenu(bookmarkB, settings);
    connect(channel, SIGNAL(received(const QCString&, const QByteArray&)),
            this, SLOT(cardMessage(const QCString &, const QByteArray &)));
    connect(bookmarkMenu, SIGNAL(bookmarkSelected(Bookmark &)), SLOT(bookmarkSelected(Bookmark &)));
    connect(bookmarkMenu, SIGNAL(documentsSelected()), SLOT(documentsSelected()));

    //load thumbnail view and listview
    init();

    fromDocument = FALSE;
    currentBookmark = 0; //documents

    if(view() == THUMBNAIL_VIEW)
        switchThumbnailView();
    else switchListView();
}

TxImage::~TxImage()
{
    settings->writeSettings();
    delete settings;
}

void TxImage::setDocument(const QString& appLnkFileName)
{
    //open an image from documents tab
    qDebug("loading document " + appLnkFileName);
    DocLnk dl(appLnkFileName);
    fromDocument = TRUE;
    showImage(&dl);

}

void TxImage::cardMessage(const QCString &msg, const QByteArray &)
{
    //new card mounted
    if(msg == "mtabChanged()")
        bookmarkMenu->load();
}

void TxImage::init(QString path)
{
    DocLnkSet *dSet=0;

    if(path.isEmpty())
    { //default to Documents
        qDebug("loading Documents...");
        dSet = new DocLnkSet();
        Global::findDocuments(dSet, "image/*");
    }else
    {
        qDebug("loading "+ path +"..." );
        dSet = new DocLnkSet(path, "image/*");
    }
    QListIterator<DocLnk> it(dSet->children());

    //load listview
    for( ; it.current(); ++it )
    {
        ImageListViewItem *item = new ImageListViewItem(listView, *it);
        listView->insertItem(item);
    }

    //load thumbnail view (use listview to sort)
    QListViewItemIterator listIt(listView);
    for( int row=0, col=0; listIt.current(); ++listIt)
    {
        //pass the listviewitem to thumbnailview to update
        //the image size in the listviewitem when the thumbnail is loaded
        //so it doesn't load the image twice
        ImageListViewItem *item = (ImageListViewItem *)listIt.current();
        thumbView->addImage(item, row, col);
        col++;
        if(col==2)
        {
            row++;
            col=0;
        }
    }

    qDebug("loading thumbnails");
    QTimer::singleShot( 0, thumbView, SLOT(loadThumbs()) );

    //select the first image
    thumbView->setFocus();
    if( listView->firstChild() )
        listView->setSelected( listView->firstChild(), TRUE );
}

void TxImage::showImage(DocLnk *d)
{
    if( settings->showImageFullScreen() )
        imageViewer->showFullScreen( d );
    else
    {
        loadingImage = d; //for delayed load
        QTimer::singleShot( 0, this, SLOT(load()) );
        imageViewer->setCaption( d->name() +" - Image Viewer");
        if(fromDocument)
            imageViewer->exec();
        else
            imageViewer->showMaximized();
    }
}

void TxImage::load()
{
    //delayed load slot
    imageViewer->load(loadingImage);
}

void TxImage::switchThumbnailView()
{
    //show the thumbnail view
    thumbViewB->setOn(TRUE);
    lView->setOn(FALSE);
    stack->raiseWidget(THUMBNAIL_VIEW);
    thumbView->setFocus();
    settings->setView( THUMBNAIL_VIEW );
}

void TxImage::switchListView()
{
    //show the list view
    stack->raiseWidget(LIST_VIEW);
    thumbViewB->setOn(FALSE);
    lView->setOn(TRUE);
    listView->setFocus();
    settings->setView( LIST_VIEW );
}

void TxImage::showBookmarkMenu()
{
    bookmarkMenu->popupMenu(bookmarkB->mapToGlobal(QPoint(0,
                          bookmarkB->height())));
}

void TxImage::bookmarkSelected(Bookmark& bm)
{
    if(currentBookmark != &bm)
    {
        //clicked bookmark isn't current bookmark
        thumbView->clear(TRUE);
        listView->clear();
        currentBookmark = &bm;
        init(bm.directory());
    }
}

void TxImage::documentsSelected()
{
    if(currentBookmark)
    {
        //current bookmark isn't documents
        thumbView->clear(TRUE);
        listView->clear();
        currentBookmark = 0; //documents
        init();
    }
}

int TxImage::view()
{
    //current view
    return settings->view();
}

void TxImage::listItemClicked(QListViewItem *listItem)
{
    if(!listItem)
        return;
    ImageListViewItem *item = (ImageListViewItem*)listItem;
    showImage( item->docLnk() );
}

void TxImage::showSettingsDialog()
{
    SettingsDialog *settingsDialog = new SettingsDialog(this);
    settingsDialog->initSettings( settings );
    if(settingsDialog->exec() == QDialog::Accepted)
    {
        //save new settings and reload bookmark menu
        settingsDialog->saveSettings();
        bookmarkMenu->load();
        settings->writeSettings();
    }

    delete settingsDialog;
}

void TxImage::setFullScreen()
{
    settings->setShowImageFullScreen( !settings->showImageFullScreen() );
    settingsPopup->setItemChecked( FULLSCREEN, settings->showImageFullScreen() );
}

void TxImage::showImageFullScreen()
{
    //show fullscreen the selected image
    if( view() == THUMBNAIL_VIEW)
    {
        ImageBox *box = thumbView->selectedImage();
        if(box)
            imageViewer->showFullScreen( box->docLnk() );
    }
    else
    {
        ImageListViewItem *item = (ImageListViewItem *)listView->selectedItem();
        if(item)
            imageViewer->showFullScreen( item->docLnk() );
    }
}

void TxImage::renameImage()
{
    DocLnk *doc=0;

    //rename the selected image
    if(view() == THUMBNAIL_VIEW) {
        ImageBox *box = thumbView->selectedImage();
        if(!box)
            return;
        doc = box->docLnk();
    }
    else {
        ImageListViewItem *item = (ImageListViewItem *)listView->selectedItem();
        if(!item)
            return;
        doc = item->docLnk();
    }

    QDialog renameDialog(this,0,TRUE);
    renameDialog.setCaption("Rename image");
    QHBoxLayout *hbox = new QHBoxLayout(&renameDialog);
    hbox->setMargin(5);
    hbox->setSpacing(5);
    QLabel *textLabel = new QLabel("New name:", &renameDialog);
    QLineEdit *renameEdit = new QLineEdit(&renameDialog);
    renameEdit->setText( doc->name() );
    hbox->addWidget(textLabel);
    hbox->addWidget(renameEdit);
    if(renameDialog.exec() == QDialog::Accepted ) {
        doc->removeLinkFile();
        doc->setName( renameEdit->text() );
        bool ok = doc->writeLink();
        if(!ok)
            qWarning("Error renaming the image");
    }
}

void TxImage::deleteImage()
{
    DocLnk *doc=0;
    ImageBox *box=0;
    ImageListViewItem *item=0;

    //delete the selected image
    if(view() == THUMBNAIL_VIEW) {
        box = thumbView->selectedImage();
        if(!box) return;
        doc = box->docLnk();
    }
    else if(view() == LIST_VIEW) {
        item = (ImageListViewItem *)listView->selectedItem();
        if(!item) return;
        doc = item->docLnk();
    }

    int ret = QMessageBox::warning(this, "Tx Image Viewer", "Are you sure do you want \n"
                                  "to delete "+doc->name()+" ?",
                                   QMessageBox::Yes, QMessageBox::No);
    if(ret == QMessageBox::Yes)
    {
        if(view() == THUMBNAIL_VIEW) {
            thumbView->deleteImage(box);
            listView->removeItem( doc );
        }
        else if(view() == LIST_VIEW) {
            listView->takeItem(item);
            thumbView->deleteImage( item->docLnk() );
        }
        doc->removeFiles();
    }

}

void TxImage::startSlideShow()
{
    docList.setAutoDelete(FALSE);
    docList.clear();

    //create a list of all images and pass it to slideshow
    QListViewItemIterator it(listView);
    for( ; it.current(); ++it)
    {
        ImageListViewItem *item = (ImageListViewItem *)it.current();
        docList.append( item->docLnk() );
    }

    //start only if there are images
    if( docList.count() )
    {
        SlideShow *slideShow= new SlideShow(&docList, settings);
        slideShow->start();
    }
}
